package org.briarproject.mailbox.core

import android.content.Context
import android.util.Log
import kotlinx.coroutines.runBlocking
import org.junit.Test
import org.sqlite2.SQLiteJDBCLoader
import org.sqlite.mc.SQLiteMCConfig
import org.sqlite.util.OSInfo
import java.io.File
import java.sql.Connection
import java.sql.DriverManager
import kotlin.test.assertEquals
import kotlin.test.assertTrue
import androidx.test.core.app.ApplicationProvider.getApplicationContext;


class DatabaseCreationTest {

    companion object {
        val LOG_TAG = "db test"
    }

    fun hasResource(path: String?): Boolean {
        return DatabaseCreationTest::class.java.getResource(path) != null
    }

    @Test
    fun createDatabase(): Unit = runBlocking {
        Class.forName("org.sqlite.JDBC")

        val ctx = getApplicationContext<Context>()
        SQLiteJDBCLoader.initialize(ctx)

        var foo = OSInfo.getNativeLibFolderPathForCurrentOS()
        Log.i(LOG_TAG, "native lib folder path: $foo")

        var sqliteNativeLibraryPath = System.getProperty("org.sqlite.lib.path")
        var sqliteNativeLibraryName = System.getProperty("org.sqlite.lib.name")
        Log.i(LOG_TAG, "native library path: $sqliteNativeLibraryPath")
        Log.i(LOG_TAG, "native library name: $sqliteNativeLibraryName")
        val mapped = System.mapLibraryName("sqlitejdbc")
        Log.i(LOG_TAG, "mapped library name: $mapped")

        if (sqliteNativeLibraryName == null) {
            sqliteNativeLibraryName = System.mapLibraryName("sqlitejdbc")
            if (sqliteNativeLibraryName != null && sqliteNativeLibraryName.endsWith(".dylib")) {
                sqliteNativeLibraryName = sqliteNativeLibraryName.replace(".dylib", ".jnilib")
            }
        }

        val packagePath =
            SQLiteJDBCLoader::class.java.getPackage().name.replace("\\.".toRegex(), "/")
        sqliteNativeLibraryPath =
            String.format("/%s/native/%s", packagePath, OSInfo.getNativeLibFolderPathForCurrentOS())
        val resource = sqliteNativeLibraryPath.toString() + "/" + sqliteNativeLibraryName
        val hasNativeLib: Boolean = hasResource(resource)
        val hasWindowsLib: Boolean = hasResource("/org/sqlite/native/Windows/x86/sqlitejdbc.dll")
        val hasLinuxFoo: Boolean = hasResource("/org/sqlite/native/Linux/android-arm/libsqlitejdbc.foo")
        val hasTor1: Boolean = hasResource("/jniLibs/armeabi-v7a/libtor.so")
        val hasTor2: Boolean = hasResource("/armeabi-v7a/libtor.so")
        val hasTor3: Boolean = hasResource("/libtor.so")
        Log.i(LOG_TAG, "resource: $resource")
        Log.i(LOG_TAG, "has native lib: $hasNativeLib")
        Log.i(LOG_TAG, "has Windows lib: $hasWindowsLib")
        Log.i(LOG_TAG, "has Linux foo: $hasLinuxFoo")
        Log.i(LOG_TAG, "has tor1: $hasTor1")
        Log.i(LOG_TAG, "has tor2: $hasTor2")
        Log.i(LOG_TAG, "has tor3: $hasTor3")

        val file: File = File.createTempFile("test", ".sqlite")
        Log.i(LOG_TAG, "file: $file")
        val connection: Connection = DriverManager.getConnection(
            "jdbc:sqlite:$file",
            SQLiteMCConfig().withKey("very-secret").toProperties()
        )

        val stmt = connection.createStatement()
        stmt.executeUpdate("CREATE TABLE students (id integer, name string)")
        stmt.executeUpdate("INSERT INTO students values (1, 'alice')")
        stmt.executeUpdate("INSERT INTO students values (2, 'bob')")

        val prep = connection.prepareStatement("SELECT * FROM students")
        val results = prep.executeQuery()
        for (i in 0..1) {
            val available = results.next()
            assertTrue(available)

            val id = results.getInt(1)
            val name = results.getString(2)
            when (i) {
                0 -> {
                    assertEquals(1, id)
                    assertEquals("alice", name)
                }
                1 -> {
                    assertEquals(2, id)
                    assertEquals("bob", name)
                }
            }
        }

        file.delete()
    }

}
